import json
import pickle

try:
    from importlib.resources import files
    from importlib.abc import Traversable
except ImportError:
    # Compatibility for Python <3.9
    from importlib_resources import files
    from importlib_resources.abc import Traversable


FILENAMES = {
    "pt3": "data/iso-639-3.tab",
    "pt2": "data/ISO-639-2_utf-8.txt",
    "pt5": "data/iso639-5.tsv",
    "retirements": "data/iso-639-3_Retirements.tab",
    "macros": "data/iso-639-3-macrolanguages.tab",
    "mapping_data": "data/iso-639.json",
    "mapping_scope": "data/iso-639_scope.json",
    "mapping_type": "data/iso-639_type.json",
    "mapping_deprecated": "data/iso-639_deprecated.json",
    "mapping_macro": "data/iso-639_macro.json",
    "list_langs": "data/iso-639_langs.pkl",
}


def get_file(file_alias: str) -> Traversable:
    """Get the path of a local data file"""
    return files(__package__).joinpath(FILENAMES[file_alias])


def load_mapping(file_alias: str) -> dict:
    """Load an ISO 639 mapping JSON file"""
    file_path = get_file(file_alias)
    try:
        with file_path.open(encoding="utf-8") as f:
            return json.load(f)
    except FileNotFoundError:
        return {}


def load_langs() -> list:
    """Load the pickled list of ISO 639 Langs"""
    file_path = get_file("list_langs")
    try:
        with file_path.open("rb") as f:
            return pickle.load(f)
    except FileNotFoundError:
        return []



from typing import Iterator

from .datafile import load_langs
from .iso639 import Lang

Lang = Lang


def iter_langs() -> Iterator[Lang]:
    """Iterate through all not deprecated ISO 639 languages

    Yields
    -------
    Lang
        Lang instances ordered alphabetically by name
    """
    sorted_langs = load_langs()

    return iter(sorted_langs)



from operator import itemgetter
from typing import Dict, List, Union

from .datafile import load_mapping
from .exceptions import DeprecatedLanguageValue, InvalidLanguageValue


class Lang(tuple):
    """Lang handles the ISO 639 series of international standards
    for language codes

    Instantiable with any ISO 639 language code or name as argument
    (case sensitive)

    ...

    Attributes
    ----------
    name : str
        ISO 639-3 reference language name or ISO 639-5 English name
    pt1 : str
        two-letter ISO 639-1 code, if there is one
    pt2b : str
        three-letter ISO 639-2 code for bibliographic applications,
        if there is one
    pt2t : str
        three-letter ISO 639-2 code for terminology applications,
        if there is one
    pt3 : str
        three-letter ISO 639-3 code, if there is one
    pt5 : str
        three-letter ISO 639-5 code, if there is one

    Examples
    --------
    >>> lg = Lang("eng")
    >>> lg
    Lang(name='English', pt1='en', pt2b='eng', pt2t='eng', pt3='eng', pt5='')
    >>> lg.name
    'English'
    """

    _tags = ("name", "pt1", "pt2b", "pt2t", "pt3", "pt5")
    _abrs = {
        "A": "Ancient",
        "C": "Constructed",
        "E": "Extinct",
        "H": "Historical",
        "I": "Individual",
        "L": "Living",
        "M": "Macrolanguage",
        "S": "Special",
    }

    _data = load_mapping("mapping_data")
    _scope = load_mapping("mapping_scope")
    _type = load_mapping("mapping_type")
    _deprecated = load_mapping("mapping_deprecated")
    _macro = load_mapping("mapping_macro")

    __slots__ = ()  # set immutability of Lang

    def __new__(cls, *args: Union[str, "Lang"], **kwargs: str):
        # validate positional argument
        if len(args) == 1:
            args_values = cls._validate_arg(args[0])
            if not args_values:
                raise InvalidLanguageValue(*args, **kwargs)
        else:
            args_values = tuple()

        # validate keyword arguments
        if kwargs:
            kwargs_values = cls._validate_kwargs(kwargs)
            if not kwargs_values:
                raise InvalidLanguageValue(*args, **kwargs)
        else:
            kwargs_values = tuple()

        # check compatiblity between positional and keyword arguments
        if args_values == kwargs_values and args_values:
            language_values = args_values
        elif args_values and not kwargs_values:
            language_values = args_values
        elif kwargs_values and not args_values:
            language_values = kwargs_values
        else:
            raise InvalidLanguageValue(*args, **kwargs)

        # instantiate as a tuple of ISO 639 language values
        return tuple.__new__(cls, language_values)

    def __repr__(self):
        chunks = ["=".join((tg, repr(getattr(self, tg)))) for tg in self._tags]

        return "Lang({})".format(", ".join(chunks))

    def __hash__(self):
        return hash(repr(self))

    def __eq__(self, other):
        return type(other) is type(self) and hash(other) == hash(self)

    def __lt__(self, other):
        return (
            type(other) is type(self) and other.name and self.name < other.name
        )

    def __getnewargs__(self):
        unpickling_args = (self.name,)

        return unpickling_args

    @property
    def name(self) -> str:
        """Gets the ISO 639 name of this language"""
        return tuple.__getitem__(self, 0)

    @property
    def pt1(self) -> str:
        """Gets the ISO 639-1 code of this language"""
        return tuple.__getitem__(self, 1)

    @property
    def pt2b(self) -> str:
        """Gets the ISO 639-2B code of this language"""
        return tuple.__getitem__(self, 2)

    @property
    def pt2t(self) -> str:
        """Gets the ISO 639-2T code of this language"""
        return tuple.__getitem__(self, 3)

    @property
    def pt3(self) -> str:
        """Gets the ISO 639-3 code of this language"""
        return tuple.__getitem__(self, 4)

    @property
    def pt5(self) -> str:
        """Gets the ISO 639-5 code of this language"""
        return tuple.__getitem__(self, 5)

    def asdict(self) -> Dict[str, str]:
        """Get ISO 639 language name and codes as a Python
        dictionary.

        Returns
        -------
        Dict[str, str]
            A dictionary containing the values of the 'name',
            'pt1', 'pt2b', 'pt2t', 'pt3' and 'pt5' attibutes.
        """

        return {tg: getattr(self, tg) for tg in self._tags}

    def scope(self) -> str:
        """Gets the ISO 639-3 scope of this language

        Returns
        -------
        str
            the ISO 639-3 scope of this language among 'Individual',
            'Macrolanguage' and 'Special'.
            None is returned by not ISO 639-3 languages.
        """
        return self._get_scope(self.pt3)

    def type(self) -> str:
        """Gets the ISO 639-3 type of this language

        Returns
        -------
        str
            the ISO 639-3 type of this language among 'Ancient',
            'Constructed', 'Extinct', 'Historical', 'Living' and
            'Special'.
            None is returned by not ISO 639-3 languages.
        """
        return self._get_type(self.pt3)

    def macro(self) -> "Lang":
        """Get the macrolanguage of this individual language

        Returns
        -------
        iso639.Lang
            the macrolanguage of this individual language, if there is one
        """
        macro_pt3 = self._get_macro(self.pt3)
        if macro_pt3:
            try:
                macro = Lang(macro_pt3)
            except InvalidLanguageValue:
                pass
            else:
                return macro

        return

    def individuals(self) -> List["Lang"]:
        """Get all individual languages of this macrolanguage

        Returns
        -------
        list of Lang
            the Lang instances of the individual languages of this
            macrolanguage, if it is one
        """
        individuals = []
        for lg in self._get_individuals(self.pt3):
            try:
                lang = Lang(lg)
            except InvalidLanguageValue:
                pass
            except DeprecatedLanguageValue as e:
                if e.change_to:
                    individuals.append(Lang(e.change_to))
            else:
                individuals.append(lang)

        return individuals

    @classmethod
    def _validate_arg(cls, arg):
        if isinstance(arg, Lang):
            return tuple(map(lambda x: getattr(arg, x), cls._tags))
        elif isinstance(arg, str):
            if len(arg) == 3 and arg.lower() == arg:
                for tg in ("pt3", "pt2b", "pt2t", "pt5"):
                    cls._assert_not_deprecated(arg)
                    values = cls._get_language_values(tg, arg)
                    if values:
                        return values
            elif len(arg) == 2 and arg.lower() == arg:
                return cls._get_language_values("pt1", arg)
            else:
                return cls._get_language_values("name", arg)

    @classmethod
    def _validate_kwargs(cls, kwargs):
        for tg in ("pt2b", "pt2t", "pt3", "pt5"):
            try:
                cls._assert_not_deprecated(kwargs[tg])
            except KeyError:
                pass

        params = set()
        for tg, lg in kwargs.items():
            if lg:
                params.add(cls._get_language_values(tg, lg))

        return params.pop() if len(params) == 1 else tuple()

    @classmethod
    def _assert_not_deprecated(cls, lang_value):
        try:
            d = cls._deprecated[lang_value]
        except KeyError:
            pass
        else:
            raise DeprecatedLanguageValue(id=lang_value, **d)

    @classmethod
    def _get_language_values(cls, tag, language_value):
        try:
            language_dict = cls._data[tag][language_value]
        except KeyError:
            language_tuple = tuple()
        else:
            language_dict[tag] = language_value
            language_tuple = itemgetter(*cls._tags)(language_dict)

        return language_tuple

    @classmethod
    def _get_scope(cls, pt3_value):
        abr = cls._scope.get(pt3_value, "")

        return cls._abrs.get(abr) if abr else None

    @classmethod
    def _get_type(cls, pt3_value):
        abr = cls._type.get(pt3_value, "")

        return cls._abrs.get(abr) if abr else None

    @classmethod
    def _get_macro(cls, pt3_value):
        return cls._macro["individual"].get(pt3_value, "")

    @classmethod
    def _get_individuals(cls, pt3_value):
        return cls._macro["macro"].get(pt3_value, [])

    @classmethod
    def _reset(cls):
        """Reload all mappings"""
        cls._data = load_mapping("mapping_data")
        cls._scope = load_mapping("mapping_scope")
        cls._type = load_mapping("mapping_type")
        cls._deprecated = load_mapping("mapping_deprecated")
        cls._macro = load_mapping("mapping_macro")



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="iso639-lang",
    version="2.2.0",
    author="L.Beaudoux",
    description="A lightweight library for the ISO 639 standard.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/LBeaudoux/iso639",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
    ],
    python_requires=">=3.7",
    package_data={"iso639": ["data/*.json", "data/*.pkl"]},
    install_requires=[
        "importlib-resources==5.12.0;python_version<'3.9'",
    ],
)



class InvalidLanguageValue(Exception):
    """Exception raised when the arguments and keyword arguments passed to
    the Lang constructor are not a valid combination of:
    - ISO 639 language name
    - ISO 639-1 language code
    - ISO 639-2B language code
    - ISO 639-2T language code
    - ISO 639-3 language code
    - ISO 639-5 language code
    """

    def __init__(self, *args, **kwargs):

        pt_msg = (
            "(*{0}, **{1}). "
            "Only valid ISO 639 language values are supported as arguments."
        )

        self.msg = pt_msg.format(args, kwargs)

        super().__init__(self.msg)


class DeprecatedLanguageValue(Exception):
    """Exception raised when the arguments and keyword arguments passed to the
    Lang constructor point to a deprecated ISO 639-3 language code
    """

    def __init__(self, *args, **kwargs):

        reasons = {
            "C": "change",
            "D": "duplicate",
            "N": "non-existent",
            "S": "split",
            "M": "merge",
        }
        reason = reasons.get(kwargs.get("reason"))
        if kwargs.get("ret_remedy"):
            remedy = "{ret_remedy}.".format(**kwargs)
        elif kwargs.get("change_to"):
            remedy = "Use [{change_to}] instead.".format(**kwargs)
        else:
            remedy = ""
        pt = (
            "As of {effective}, [{id}] for {name} is deprecated "
            "due to {0}. {1}"
        )

        self.msg = pt.format(reason, remedy, **kwargs)
        self.name = kwargs.get("name", "")
        self.change_to = kwargs.get("change_to", "")
        self.ret_remedy = kwargs.get("remedy", "")
        self.effective = kwargs.get("effective", "")

        super().__init__(self.msg)
